from argparse import ArgumentParser
from functools import partial
from pandas import DataFrame
import sys
import warnings

from opa_tools.benchmarking import (
    get_environ_df,
    get_request_df,
)
from opa_tools.utils import (
    add_date_parsing_args,
    parse_date_interval,
)
from .utils import format_dataframe


def _get_parser():
    parser = ArgumentParser(
        prog="bch",
        description="""
        Benchmarking CLI for comparing request and environment performance over time.

        This tool allows you to extract tabular performance data for:
        - Request-level benchmarking (`req`)
        - Environment-level performance (`env`)
        - Column discovery (`col`)

        You can filter by date or date intervals, and format output for integration with other tools.
        """
    )

    subparsers = parser.add_subparsers(
        title="Service",
        dest="service",
        metavar="COMMAND",
        help="Benchmarking subcommand to run. Use `bch <COMMAND> --help` for details.",
    )

    # req: Show request-level benchmark data
    parser_req = subparsers.add_parser(
        "req",
        help="Show request-level benchmark data in tabular form",
        description="""
        Retrieve request performance data (e.g., durations, errors, identifiers)
        for one or more dates or date ranges.
        """
    )
    _add_table_arguments(parser_req)

    # env: Show environment-level benchmark data
    parser_env = subparsers.add_parser(
        "env",
        help="Show environment-level benchmark data (e.g., runtime metrics)",
        description="""
        Retrieve environment-level benchmark data (e.g., CPU/memory usage, runtime)
        filtered by dates or intervals.
        """
    )
    _add_table_arguments(parser_env)

    # col: List columns in either req or env tables
    parser_col = subparsers.add_parser(
        "col",
        help="List available benchmark columns for 'req' or 'env'",
        description="Prints a sorted list of all available columns in the specified dataset.",
    )
    parser_col.add_argument(
        "which",
        choices=("req", "env"),
        type=str,
        default="req",
        help="Which dataset to inspect for column names. Choose 'req' or 'env'.",
    )

    return parser


def _add_table_arguments(parser):
    parser.add_argument(
        "-c", "--columns",
        action="store",
        nargs="*",
        default=[],
        metavar="COLUMN",
        help="Specify which columns to display (e.g. 'duration', 'error'). Always includes 'date' and 'bname'."
    )

    add_date_parsing_args(parser)  # adds: --dates and --interval

    parser.add_argument(
        "-f", "--format",
        action="store",
        type=str,
        choices=["default", "fwf"] + [
            method.lstrip("to_")
            for method in dir(DataFrame) if method.startswith("to_")
        ],
        default="default",
        help="Output format: default, fwf (fixed-width), or any pandas-supported format like csv, json, markdown.",
    )

    parser.add_argument(
        "-t", "--tablefmt",
        action="store",
        type=str,
        metavar="TABULAR_TABLEFMT",
        default="plain",
        help="Optional: tabulate-style table format for text output (e.g., 'grid', 'fancy_grid', 'plain')."
    )


def _validate_columns(requested, existing):
    def force_column(lst, col):
        if col in existing:
            if col in lst:
                return
            lst.insert(0, col)
    columns = requested.copy()
    force_column(columns, "date")
    force_column(columns, "bname")
    wrong_columns = list(filter(lambda x: x not in existing, columns))
    if wrong_columns:
        warnings.warn(
            f"Column(s) {', '.join(wrong_columns)} skipped."
        )
        columns = [column for column in columns if column not in wrong_columns]
    return columns

def _table_service(args, fun):
    if not args.interval:
        df = fun(dates=args.dates)
    else:
        df = fun(date_interval=parse_date_interval(args.dates))
    if args.columns:
        columns = _validate_columns(args.columns, df.columns)
        df = df[columns]
    print(format_dataframe(df, args.format, args.tablefmt))

def _column_service(args):
    df = (
        get_request_df()
        if args.which == "req"
        else get_environ_df()
    )
    print("\n".join(sorted(df.columns)))


_AVAILABLE_SERVICES = {
    "req": partial(_table_service, fun=get_request_df),
    "env": partial(_table_service, fun=get_environ_df),
    "col": partial(_column_service),
}

def bch(argv=None):
    if argv is None:
        argv = sys.argv[1:]

    parser = _get_parser()
    args = parser.parse_args(argv)
    service = args.service
    if not service:
        parser.print_help()
        sys.exit(1)
    _AVAILABLE_SERVICES[service](args)

