"""
This module provides tools for generating new request directories from templates.
It supports creating request templates (`create_template`) and populating request
directories with actual request data (`generate_requests`) based on YAML specifications.

Main functionalities:
- Wraps and flattens request data for template creation.
- Expands template combinations into individual request folders.
"""

from datetime import datetime
from math import floor, log10
from typing import List, Union
import shutil
import yaml

from opa_tools.config import REQUEST_DIR
from .retrieval import get_request
from .transformation import product, unflatten, wrap
from .utils import (
    get_request_location,
    normalize_request_identifier,
    _normalize_request_month,
    ALLOWED_KINDS,
    validate_request_kind,
)


def _write_requests(requests, grequest_dir, force):
    kinds = ALLOWED_KINDS
    with open(grequest_dir / "date.txt", "w") as wf:
        wf.write(datetime.now().isoformat())
    n_digits = floor(log10(len(requests))) + 1
    for i, request in enumerate(requests, start=1):
        request_dir = REQUEST_DIR / grequest_dir / f"{i:0{n_digits}d}"
        if request_dir.exists() and force:
            shutil.rmtree(request_dir)
        try:
            request_dir.mkdir(exist_ok=False)
        except FileExistsError as exc:
            raise FileExistsError(
                f"Directory {request_dir} exists. Use the `--force` argument to overwrite."
            ) from exc
        for kind, request_kind in zip(kinds, unflatten(request, kinds)):
            with open(request_dir / f"{kind}.yml", "w") as wf:
                yaml.dump(request_kind, wf)


def create_template(
    src: Union[str, List[str]],
    dst: str,
) -> None:
    name, _, month = normalize_request_identifier(dst, None, None)
    new_dir = REQUEST_DIR / month / name
    if new_dir.exists():
        raise IsADirectoryError(
            f"Directory '{new_dir}' exists. Choose a different dst."
        )
    if isinstance(src, str):
        src = [src]
    generator_request = wrap([ # pyright: ignore
        get_request(element, flat=False, single_as_list=False) for element in src
    ])
    new_dir.mkdir()
    new_path = new_dir / "grequest.yaml"
    with open(new_path, "w") as wf:
        yaml.dump(generator_request, wf)
    print(f"File {new_path} created.")


def generate_requests(
    identifier: Union[str, int],
    force: bool = False,
) -> None:
    name, _, month = normalize_request_identifier(identifier, None, None)
    grequest_path = (
        get_request_location(name, month=month, no_exist_ok=True)
        / "grequest.yaml"
    )
    if not grequest_path.exists():
        raise FileNotFoundError(
            f"Generating request attempted to be found in '{grequest_path}', but none found. "
            "Have you run `create_template`?"
        )
    with open(grequest_path, "r") as rf:
        generating_request = yaml.safe_load(rf)
    requests = product(generating_request)
    _write_requests(requests, grequest_path.parent, force)

