from argparse import ArgumentParser, RawDescriptionHelpFormatter
from concurrent.futures import as_completed, ProcessPoolExecutor
from datetime import datetime
from pathlib import Path
import time
import shelve
from typing import TYPE_CHECKING, Sequence
import os
import sys
import yaml

import numpy as np
from numpy.typing import NDArray

from crick import TDigest
try:
    from crick import TDigestArray
except ImportError:
    TDigestArray = None

# Check if we are under an MPI environment
MPI_LAUNCHED = any(var in os.environ for var in [
    "OMPI_COMM_WORLD_SIZE",  # OpenMPI
    "PMI_SIZE",              # MPICH / IntelMPI
    "MPI_LOCALNRANKS",       # Some others
])

if MPI_LAUNCHED:
    try:
        from mpi4py import MPI
        MPI_COMM = MPI.COMM_WORLD
        MPI_RANK = MPI_COMM.Get_rank()
        MPI_SIZE = MPI_COMM.Get_size()
    except ImportError:
        MPI, MPI_COMM, MPI_RANK, MPI_SIZE = None, None, None, None
else:
    MPI, MPI_COMM, MPI_RANK, MPI_SIZE = None, None, None, None

if TYPE_CHECKING:
    from one_pass.opa import Opa


DEFAULT_N_DIGESTS = 40000
DEFAULT_INIT_SIZE = 1000
DEFAULT_WEIGHT = 1

METHODS = [
    "serial",
    "mp",
    "mpi",
    "tda",
]


def _get_parser() -> ArgumentParser:
    parser = ArgumentParser(
        description="Program to perform TDigest performance.",
        formatter_class=RawDescriptionHelpFormatter,
    )
    parser.add_argument("method", choices=METHODS, action="store")
    parser.add_argument("-k", "--workers", type=int, action="store", default=1)
    parser.add_argument(
        "-n", "--ndigests", type=int, action="store", default=DEFAULT_N_DIGESTS
    )
    parser.add_argument(
        "-i", "--init-size", type=int, action="store", default=DEFAULT_INIT_SIZE
    )
    parser.add_argument(
        "-w", "--weight", type=int, action="store", default=DEFAULT_WEIGHT
    )
    return parser


def benchmark(fun):
    def wrapper(*args, **kwargs):
        t0 = time.time()
        res = fun(*args, **kwargs)
        t1 = time.time()
        return res, t1 - t0

    return wrapper


def init_digests(n: int, init_size: int = 0):
    digest_list = np.array([None] * n, dtype=object)

    if MPI is None or MPI_SIZE == 1:
        for j in range(n):
            digest_list[j] = TDigest(compression=1)
            if init_size:
                digest_list[j].update(np.random.normal(size=init_size))
    else:
        _mpi_init_digests(n, init_size, digest_list)

    return digest_list


def _mpi_init_digests(size: int, init_size: int, digest_list: Sequence[object]):
    chunk_size = size // MPI_SIZE

    # Scatter the work among all processes
    start_index = MPI_RANK * chunk_size
    if MPI_RANK == MPI_SIZE - 1:  # Last process handles the remainder
        end_index = size
    else:
        end_index = (MPI_RANK + 1) * chunk_size

    for j in range(start_index, end_index):
        digest_list[j] = TDigest(compression=1)
        if init_size:
            digest_list[j].update(np.random.normal(size=init_size))


def _get_input_and_output_structure(digests: Sequence[TDigest], size: int, weight: int):
    if weight == 1:
        values = np.random.normal(size=size)
        # Prepare input for parallel processing
        inputs = [(j, values[j], digests[j]) for j in range(size)]
        outputs = [None] * size

    else:
        values = np.random.normal(size=(weight, size))
        # Prepare inputs for parallel update
        inputs = [(j, values[:, j], digests[j]) for j in size]
        outputs = [None] * size

    return inputs, outputs


def _update_single_digest(j, value, digest):
    """Helper function to update a single digest."""
    digest.update(value)
    return j, digest


@benchmark
def update_mp(inputs: Sequence[object], outputs: Sequence[object], workers: int):
    """Parallel version: Updates the digest for each grid point. Uses multiprocessing."""

    with ProcessPoolExecutor(max_workers=workers) as executor:
        futures = {
            executor.submit(_update_single_digest, *args): args[0] for args in inputs
        }
        for future in as_completed(futures):
            j, updated_digest = future.result()
            outputs[j] = updated_digest

    return outputs


@benchmark
def update_mpi(inputs: Sequence[TDigest], outputs: Sequence[object]):
    """Parallel version: Updates the digest for each grid point. Uses MPI."""

    num_items = len(inputs)
    chunk_size = num_items // MPI_SIZE

    # Scatter the work among all processes
    start_index = MPI_RANK * chunk_size
    if MPI_RANK == MPI_SIZE - 1:  # Last process handles the remainder
        end_index = num_items
    else:
        end_index = (MPI_RANK + 1) * chunk_size

    local_inputs = inputs[start_index:end_index]  # Slice for local processing

    local_outputs = [None] * len(local_inputs)
    for i, args in enumerate(local_inputs):
        local_outputs[i] = _update_single_digest(*args)

    gathered_outputs = MPI_COMM.gather(local_outputs, root=0)

    # Root process has MPI_RANK 0
    if MPI_RANK == 0:
        flat_outputs = [item for sublist in gathered_outputs for item in sublist]
        for idx, updated_digest in flat_outputs:
            outputs[idx] = updated_digest

    # Other processes keep only their local outputs
    else:
        outputs[start_index:end_index] = local_outputs


@benchmark
def update_serial(inputs: Sequence[object], outputs: Sequence[object]):
    for args in inputs:
        j, value = _update_single_digest(*args)
        outputs[j] = value


@benchmark
def update_tda(tda: TDigestArray, data: NDArray):
    tda.update(data)


def compute(method: str, ndigests: int, init_size: int, workers: int):

    if method == "serial":
        digests = init_digests(ndigests, init_size)
        inputs, outputs = _get_input_and_output_structure(digests, ndigests, 1)
        _, t = update_serial(inputs, outputs)

    elif method == "mp":
        digests = init_digests(ndigests, init_size)
        inputs, outputs = _get_input_and_output_structure(digests, ndigests, 1)
        _, t = update_mp(inputs, outputs, workers)

    elif method == "mpi":
        digests = init_digests(ndigests, init_size)
        inputs, outputs = _get_input_and_output_structure(digests, ndigests, 1)
        _, t = update_mpi(inputs, outputs)

    elif method == "tda":
        if TDigestArray is None:
            raise ImportError(
                "Forked version of crick implementing TDigestArray should be installed."
            )
        tda = TDigestArray(ndigests)
        tda.update(np.random.normal(size=(init_size, ndigests)))
        tda_update_data = np.random.normal(size=(1, ndigests))
        _, t = update_tda(tda, tda_update_data)

    else:
        raise ValueError(f"Wrong method: {method}.")

    return t


def write_stats(path: Path, **kwargs):
    try:
        path.mkdir(parents=True, exist_ok=False)
    except FileExistsError as exc:
        raise FileExistsError(
            "Benchmark directory already exists!"
        ) from exc

    for key, value in kwargs.items():
        with open(path / f"{key}.yaml", "w", encoding="utf-8") as wf:
            yaml.safe_dump(value, wf)

def main():

    args = _get_parser().parse_args()
    now = datetime.now()

    try:
        benchmark_dir = Path(os.environ["BENCHMARK_DIR"]).resolve()
        benchmark_dir /= "crick"
        benchmark_dir /= now.strftime(r"%Y%m%d_%H%M%S_%f")
    except KeyError as exc:
        raise EnvironmentError(
            "Please set environment variable 'BENCHMARK_DIR'."
        ) from exc

    benchmark_time = compute(
        method=args.method,
        ndigests=args.ndigests,
        init_size=args.init_size,
        workers=args.workers,
    )

    if MPI is not None and MPI_SIZE > 1 and MPI_RANK > 0:
        # MPI installed, MPI enabled, non-root process.
        sys.exit(0)

    if MPI is not None and MPI_SIZE > 1:
        workers = MPI_SIZE
    else:
        workers = args.workers

    write_stats(
        benchmark_dir, 
        datetime=now.isoformat(" "),
        args={
            "method": args.method,
            "ndigests": args.ndigests,
            "init_size": args.init_size,
            "workers": workers,
        },
        argv=list(sys.argv),
        environ=dict(os.environ),
        benchmark=benchmark_time,
    )


if __name__ == "__main__":
    main()
