from datetime import datetime
import pytest

from opa_tools.request.utils import (
    ALLOWED_KINDS,
    normalize_request_identifier,
    validate_request_kind,
)

yymm = datetime.now().strftime("%y%m")
yy = yymm[0:2]

@pytest.mark.parametrize("input, expected", [
    (("001", None), ("001", yymm)),
    ((1, None), ("001", yymm)),
    (("05/1", None), ("001", f"{yy}05")),
    (("05/001", None), ("001", f"{yy}05")),
    (("1809/1", None), ("001", "1809")),
    (("1809/001", None), ("001", "1809")),
    (("001", "1904"), ("001", "1904")),
    ((1, "1904"), ("001", "1904")),
    (("001", "03"), ("001", f"{yy}03")),
    ((1, "03"), ("001", f"{yy}03")),
])
def test_identifier_normalization(input, expected):
    i_identifier, i_month = input
    e_name, e_month = expected
    o_name, _, o_month = normalize_request_identifier(
        i_identifier, month=i_month
    )
    assert (o_name, o_month) == (e_name, e_month)

@pytest.mark.parametrize("input, expected", [
    (("001", 1, None), ("001", "1", yymm)),
    (("001", "1", None), ("001", "1", yymm)),
    (("05/1.1", None, None), ("001", "1", f"{yy}05")),
    (("05/01.1", None, None), ("001", "1", f"{yy}05")),
    (("05/001.01", None, None), ("001", "1", f"{yy}05")),
    (("05/1.01", None, None), ("001", "1", f"{yy}05")),
    (("1809/1.2", None, None), ("001", "2", "1809")),
    (("1809/001.3", None, None), ("001", "3", "1809")),
    (("05/1", "1", None), ("001", "1", f"{yy}05")),
    (("05/01", "1", None), ("001", "1", f"{yy}05")),
    (("05/1", "01", None), ("001", "1", f"{yy}05")),
    (("05/001", "01", None), ("001", "1", f"{yy}05")),
    (("1809/1", "2", None), ("001", "2", "1809")),
    (("1809/001", "3", None), ("001",  "3", "1809")),
    (("001.6", None, "1904"), ("001", "6", "1904")),
    (("001", "6", "1904"), ("001", "6", "1904")),
    ((1, 6, "1904"), ("001", "6", "1904")),
    ((1.6, None, "1904"), ("001", "6", "1904")),
    (("001.9", None, "03"), ("001", "9", f"{yy}03")),
    (("001", "9", "03"), ("001", "9", f"{yy}03")),
    ((1.5, None, "03"), ("001", "5", f"{yy}03")),
])
def test_subrequest_normalization(input, expected):
    i_identifier, i_subrequest, i_month = input
    e_name, e_subrequest, e_month = expected
    if isinstance(i_identifier, float):
        with pytest.raises(NotImplementedError):
            normalize_request_identifier(
                i_identifier, subrequest=i_subrequest, month=i_month
            )
        return
    output = normalize_request_identifier(
        i_identifier, subrequest=i_subrequest, month=i_month
    )
    o_name, o_subrequest, o_month = output
    assert (o_name, o_subrequest, o_month) == (e_name, e_subrequest, e_month)

@pytest.mark.parametrize("test_kind", [
    "gsv", "opa", "par", "xxx", "will_cause_error"
])
def test_request_type(test_kind):
    if test_kind not in ALLOWED_KINDS:
        with pytest.raises(ValueError):
            validate_request_kind(test_kind)
        return
    validate_request_kind(test_kind)
